require "util"
--local logtable = {}
local talkative = true
local newfile = true

function print(text, ...)
--	if true then return end
	if newfile then game.write_file("fws.log","new log",false) newfile = false end
	if type(text)~="string" then text = serpent.block(text) end
	if ... then text = string.format(text,...) end
	text = text.."\n\r"
	if talkative then for _,p in pairs(game.players) do
		p.print(text)
	end end
	game.write_file("fws.log",text,true)
	--[[
	logtable[#logtable]=text
	game.write_file("fws.log",table.concat(logtable,"\n"))
	--]]
end


--[[
script.on_event(defines.events.on_player_created, function(e)
	local p = game.players[e.player_index]
	p.print("FWS test mode: giving free schtuffs")
	p.insert("raw-wood")	--test
	p.insert("spores")	--test
end)
--]]

local rand, sqrt, floor, ceil = math.random, math.sqrt, math.floor, math.ceil
local substr,char,ascii = string.sub,string.char,string.byte


local DEFAULT_CONDITIONS = {water=.5,water_range=.3,temp=20,temp_range=10}

local DEATH_TABLE, DEATH_THRESHOLD
local queue, sporelist, tile_fertility, growlist

function alias()
	DEATH_TABLE = global.DEATH_TABLE
	DEATH_THRESHOLD = global.DEATH_THRESHOLD
	
	queue = global.queue
	sporelist = global.sporelist
	tile_fertility = global.tile_fertility
	growlist = global.growlist
end

local queuematch = {}

function QueueTick(entity,ticks)
	if not entity or not ticks then return end
	ticks = ceil(game.tick + (ticks or 1))
	if queue[ticks] then
		table.insert(queue[ticks],entity)
	else
		queue[ticks]={entity}
	end
	queuematch[ticks]={game.tick,entity.position}
	--print(entity.name.."("..serpent.line(entity.position,{nocode=true}).."} for tick "..ticks)
end

function decodenum(letter)
	local a = ascii(letter)
	if a < 65 then
		if a < 49 then return 1 else return a-48 end
	else
		if a > 89 then return 35 else return a-55 end
	end
end

function encodenum(number)
	if number>9 then
		if number>35 then return "Z" end
		return char(number+55)
	else
		if number<2 then return "1" end
		return char(number+48)
	end
end

--- 1234567890123456789
---"blah-growing-X-of-Y"
-- -9876543210987654321

function findgrowables()
	local namelist = {}
	local prototypes=game.entity_prototypes
	local items = game.item_prototypes
	local root
	local temp
	
	local solar_index = 700/9 --J per tiletick (day avg)
	--[[
	for name,proto in pairs(game.entity_prototypes) do
		if proto.type=="solar-panel" then
			temp = proto.collision_box
			temp = (temp.right_bottom.x-temp.left_top.x) * (temp.right_bottom.y-temp.left_top.y)
			temp = 123456789 /60*.7/temp
			if temp > solar_index then solar_index=temp end
		end
	end
	--]]
	
	for name,proto in pairs(game.tile_prototypes) do
		temp = {}
		global.tile_fertility[name] = temp
		temp.absorb = proto.emissions_per_tick
		if temp.absorb<0 then temp.absorb = 0 end
		
		root = proto.autoplace_specification
		
		temp.temp = 0
		temp.water = 0
		if root then
			for _,p in pairs(root.peaks) do
				if p.temperature_optimal then temp.temp = p.temperature_optimal end
				if p.water_optimal then temp.water = p.water_optimal end
			end
		end
	end
	
	for name in pairs(prototypes) do
		if substr(name,-15,-7)=="-growing-" and substr(name,-5,-2)=="-of-" then
			root=substr(name,1,-16)
			if prototypes[root] then
				if not namelist[root] then namelist[root]={} end
				namelist[root][decodenum(substr(name,-6,-6))]=name
			end
		end
	end
	
	--print(namelist)

	

	--growlist, sporelist ={},{}
	for k in pairs(growlist) do growlist[k]=nil end
	for k in pairs(sporelist) do sporelist[k]=nil end
	for base,list in pairs(namelist) do
--		game.print(base)
		root = nil
		if list[1] then growlist[base]={spawns=list[1]} end
		if string.find(base,"tree") or string.find(base,"coral") then
			if prototypes[base].autoplace_specification and prototypes[base].mineable_properties.minable then
--				game.print("minable")
				temp = 0
				for _,p in pairs(prototypes[base].mineable_properties.products) do
--					game.print(serpent.line(p))
					if p.name == "raw-wood" then
						sporelist[base]=(sporelist[base] or 0) + (p.amount_max or p.amount) * items[p.name].fuel_value
					end
				end
				
			end
		end
		
		if string.find(base,"tree") then DEATH_TABLE[base]="dead-tree" end
		
		if prototypes[base].autoplace_specification then
			for _,peak in pairs(prototypes[base].autoplace_specification.peaks) do
				if peak.temperature_optimal then
					if not root then root = {} end
					root.temp = peak.temperature_optimal
					root.temp_range = peak.temperature_range
				end
				if peak.water_optimal then
					if not root then root = {} end
					root.water = peak.water_optimal
					root.water_range = peak.water_range
				end
			end
			if root then growlist[base].conditions = root end
		end
		
		--print("%s has %i stages and fuelgrowth %i; phase ticksize %f",base,#namelist[base], sporelist[base] or 0, (sporelist[base] or 16000000)/#namelist[base]/solar_index )
		
		for stage,name in pairs(list) do
			growlist[name] = {	next=list[stage+1] or base,
								ticks=ceil((sporelist[base] or 16000000)/#namelist[base]/solar_index), max_health = prototypes[name].max_health,
								base=base}
		end
	end
	--print(sporelist)
	--print(growlist)
end

script.on_load(alias)
script.on_init(function()
	if not global.DEATH_TABLE then global.DEATH_TABLE = {} end
	if not global.queue then global.queue = {} end
	if not global.sporelist then global.sporelist = {} end
	if not global.tile_fertility then global.tile_fertility = {} end
	if not global.growlist then global.growlist = {} end
	
	if not global.DEATH_THRESHOLD then global.DEATH_THRESHOLD = game.map_settings.pollution.min_pollution_to_damage_trees end
	if not global.DAYLENGTH then global.DAYLENGTH = 25000 end
	alias()
	findgrowables()
end)
script.on_configuration_changed(function()
	local area = {left_top={},right_bottom={}}
	global.DEATH_THRESHOLD = game.map_settings.pollution.min_pollution_to_damage_trees
	alias()
	findgrowables()
	for _,surf in pairs(game.surfaces) do
		area.surface = surf
		for chunk in surf.get_chunks() do
			area.left_top.x = chunk.x*32
			area.left_top.y = chunk.y*32
			area.right_bottom.x = chunk.x*32+32
			area.right_bottom.y = chunk.y*32+32
			--print("callingsetwildlife on %s",serpent.line(area))
			setwildlife(area, false)
		end
	end
end)

--[[
growlist[entityname]
	.spawns
and/or
	.ticks
	.next

--]]

--  {...}"-growing-S-of-M" (123456789ABCDEF max 35="Z")
-- sub(name,

remote.add_interface("fws",
{
	register = function(adds)
		for k,v in pairs(adds) do
			growlist[k]=util.table.deepcopy(v)
		end
	end,
	growqueue = function()
		print("FWS grow queue")
		for k,v in pairs(queue) do
			for _,e in pairs(v) do
--				print(serpent.line(e))
				if e.valid then
					if e.type ~= "resource" then
						print("%i: %s (%0.2f,%0.2f)",k,e.name,e.position.x,e.position.y) --serpent.line(v))--string.format("%i: %s",k,#v))
					end
	--				return
				end
			end 
		end
	end,
	chunkqueue = function()
		print("FWS chunk queue")
		for k,v in pairs(queue) do
			for _,e in pairs(v) do
				if e.valid then
					if e.type == "resource" then
						print("%i: (%i,%i), fertility %i",k,e.position.x/32-.5,e.position.y/32-.5,e.amount) --serpent.line(v))--string.format("%i: %s",k,#v))
					end
				end
			end 
		end
	end,
	fitness = function()
		for k,v in pairs(growlist) do
			if v.conditions then
				print(string.format("%s: %f,%f\n",k,v.conditions.temp, v.conditions.water))
			end
		end
	end,
	renew = findgrowables,
	growtable = function()
		print(serpent.block(growlist))
	end,
	talk = function() talkative = not talkative end,
})

function setwildlife(area, purge)
	local counter = area.surface.find_entities_filtered{area=area,name="counter-tree"}
	if purge then for _,e in pairs(counter) do e.destroy() end
	else if #counter>0 then return end end
	
	local treecount = area.surface.count_entities_filtered{area=area,type="tree"}+1

	counter = area.surface.create_entity{name="counter-tree",position={area.left_top.x+16,area.left_top.y+16}}
	counter.amount=treecount
	--print("initialized chunk (%i,%i)",area.left_top.x, area.left_top.y)
	QueueTick(counter,rand()*300+120)
	return counter
end
 
function grow_tree_at(surface,pos)
	local weights,sum,bit={},0
	for candidate in pairs(sporelist) do
		bit = GrowFitness(growlist[candidate],surface,pos) -- + .5
		--bit = bit * bit - .25
		weights[candidate]=bit
		sum = sum + bit
		--print(candidate.." "..bit)
	end
	if sum==0 then return end
	sum = sum * rand()
	for c,w in pairs(weights) do
		if sum < w then
			bit = surface.find_non_colliding_position(c,pos,.6,.1)
			if bit then
				bit=surface.create_entity{name=growlist[c].spawns,position=bit}
				--bit.health = bit.health * w
				QueueTick(bit,(growlist[bit.name].ticks or 50000)*(1.4-w*rand()/2 ))
				--print("chosen "..c)
			else
				--print("failed to place "..c.." (this should not happen)")
			end
			return
		else
			sum = sum - w
		end
	end
end

script.on_event(defines.events.on_trigger_created_entity, function(e)
	if e.entity.name == "spore-trigger" then
		local pos = e.entity.position
		pos.x=pos.x-.5+rand()
		pos.y=pos.y-.5+rand()
		grow_tree_at(e.entity.surface,pos)
	end
end)

script.on_event(defines.events.on_chunk_generated, function(event) event.area.surface = event.surface; setwildlife(event.area, true)
--for _,d in pairs(event.surface.find_entities_filtered{area=event.area,type="decorative"}) do d.destroy() end
 end)

script.on_event(defines.events.on_tick, function(e)

	local t=e.tick
	
	local event = queue[t]
	if not event then return end
	
	for _,queued in pairs(event) do
		if queued.valid then
			if queued.type == "tree" then
				GrowTree(queued)
			elseif queued.type == "resource" then
				VisitChunk(queued)
			else
				--reserved for future/error
			end
		else
			--print(t.."invalid"..serpent.line(queuematch[t]))
		end
	end
	
	queue[t] = nil
	queuematch[t] = nil
end)

function GrowFitness(entry,surface,pos)
	if not entry.base then entry = entry.conditions or DEFAULT_CONDITIONS
	else entry = entry.base.conditions or DEFAULT_CONDITIONS end
	
	local tile = tile_fertility[surface.get_tile(pos.x,pos.y).name]
	local fit = tile.absorb
	
	--TODO: hard tile overrides (for specialty farming etc)
	
	if fit==0 then return 0 end
	--print(tile.absorb)
	
	--HACK until tile data is back (TODO)
	
	local tf,wf = (tile.temp - entry.temp)/entry.temp_range, (tile.water - entry.water) / entry.water_range

	if tf<0 then tf=-tf end
	if tf<fit then tf=0 end
	if wf<0 then wf=-wf end
	if wf<fit then wf=0 end
	
	if tf>1 or wf>1 then
		return .5/(tf+wf)
	else
		return 1-(tf*wf)/2
	end
	
end

--anchor is either counter/index for chunk, or tree
function VisitChunk(anchor)
	local area = {{anchor.position.x-16,anchor.position.y-16},{anchor.position.x+16,anchor.position.y+16}}
	local surface = anchor.surface

--	print("Visiting "..floor(anchor.position.x/32) ..","..floor(anchor.position.y/32))
	
	local r = rand()
	
	local pollution = surface.get_pollution(anchor.position)

	local o=anchor.amount
	local thinglist = surface.find_entities_filtered{area=area,type="tree"}
	local count = #thinglist
	local e
	local fit1,fit2
	
	--print("anchor position is "..serpent.line(anchor.position))
		--print("Visiting %i,%i: %i/%i trees. (chance is %f, r=%f).",floor(anchor.position.x/32),floor(anchor.position.y/32),count,o,3*count*(o-count)/(o*o),r)
	if count>0 and count~=o and r < 3*count*(o-count)/(o*o) then
		--start seedling
		--print("Visiting %i,%i: %i/%i trees. (chance is %f, r=%f).",floor(anchor.position.x/32),floor(anchor.position.y/32),count,o,3*count*(o-count)/(o*o),r)
		r = rand()
		e = thinglist[floor(r*(count-.1))+1]
		o = growlist[e.name]
		--print("o is %s",e.name)
		if o and o.spawns then
			--print("Spawning %s from %s.",o.spawns,e.name)
			o = o.spawns
			local p1,p2, f1,f2 =	surface.find_non_colliding_position(e.name,{e.position.x+r*5,e.position.y+rand()*5},12,.1),
									surface.find_non_colliding_position(e.name,{e.position.x+rand()*5,e.position.y+r*5},12,.1),
									0, 0
			if p1 then f1=GrowFitness(o,surface,p1) end
			if p2 then f2=GrowFitness(o,surface,p2) end
			
			if (f1+f2)>0 then
				if f2 > f1 then
					e=surface.create_entity{name=o,position=p2}
					QueueTick(e,(growlist[e.name].ticks or 50000)*(.75+rand()/2))
					--print("growing "..serpent.line(e.position))
					--GrowTree{type=e.name,stage=1,pos=p2}
				else
					e=surface.create_entity{name=o,position=p1}
					QueueTick(e,(growlist[e.name].ticks or 50000)*(.75+rand()/2))
					--print("growing "..serpent.line(e.position))
					--GrowTree{type=e.name,stage=1,pos=p1}
				end
			end
		end
	end

	if pollution>DEATH_THRESHOLD and count>0 then
		o=floor((pollution/DEATH_THRESHOLD)*r)
		while o>1 do
		o=o*.7
		--if r*e>1 then --todo maybe change formula?
			e=thinglist[floor(rand()*(count-.1))+1]
			if e.valid then
				r=DEATH_TABLE[e.name]
				if r then
					--print(e.name.." died, became "..r)
					surface.create_entity{name=r, position=e.position, force=game.forces.neutral}
					e.destroy()
				else
					--print(e.name.." died.")
					e.destroy()
				end
			end
		end
		--thinglist[floor(r*(count-.1))+1]
		r=rand()
	end

	thinglist = surface.find_entities_filtered{area=area,type="fish"}
	count = #thinglist
	if count>0 then
		e = thinglist[floor(r*(count-.1))+1]
		o=surface.count_entities_filtered{area={{e.position.x-9,e.position.y-9},{e.position.x+9,e.position.y+9}},name=e.name}
		r=rand()
		if o>1 and o<5 then
			if o==3 then r = r / 2 end
			if r<.05 then	--10% chance at 3, 5% at 2 or 4
				surface.create_entity{name= e.name, position= e.position, force= game.forces.neutral}
			end
			r=rand()
		end
	end
	QueueTick(anchor,global.DAYLENGTH*(.75+rand()/2))
end

function GrowTree(growable)--{type="",pos={x,y},stage=n,entity=0x}
	--if growable then print("isagrowable") end
	--if growable.valid then print("isvalid") end

	if not (growable and growable.valid) then print("Error: invalid tree passed to growth routine") return end
	local growdata, surface, force = growlist[growable.name], growable.surface, growable.force
	if not growdata then return end	--shouldn't happen, but sanity check
	
	if growdata.next then
		local pos = growable.position
		local area = {{pos.x-16,pos.y-16},{pos.x+16,pos.y+16}}
		
		local health = growable.health/growdata.max_health
		local variation = growable.graphics_variation
		growable.destroy()
		
		e=surface.create_entity{name=growdata.next,position=pos,force=force}
		e.health = e.health*health
		if variation then e.graphics_variation = variation end
		
		health = GrowFitness(growdata.base,surface,pos)* (DEATH_THRESHOLD/(surface.get_pollution(pos)+DEATH_THRESHOLD))	--growfitness and pollution penalty
		local crowd = surface.find_entities_filtered{area=area,name="counter-tree"}[1]

		crowd = sqrt((surface.count_entities_filtered{area=area,type="tree"}+1)/(crowd and crowd.amount+1 or 2))/.875
		--print("%i/%i crowding penalty %.2f",c,l,crowd)
		QueueTick(e,(growdata.ticks or 50000)*(1.5-health*rand()/2)*crowd)	--time * 1 .. 1.5
	end

end

